@@ -19,4 +19,4 @@
 extern crate log;
 
 mod agent;
-pub use agent::ocall_handle_file_request;
+pub use agent::{handle_file_request, ocall_handle_file_request};
