@@ -34,13 +34,16 @@ pub use teaclave_proto::teaclave_authentication_service::{
 pub use teaclave_proto::teaclave_frontend_service::GetFunctionResponse as Function;
 pub use teaclave_proto::teaclave_frontend_service::{
     ApproveTaskRequest, ApproveTaskResponse, AssignDataRequest, AssignDataResponse,
-    CreateTaskRequest, CreateTaskResponse, GetFunctionRequest, GetFunctionResponse, GetTaskRequest,
-    GetTaskResponse, InvokeTaskRequest, InvokeTaskResponse, RegisterFunctionRequest,
+    CancelTaskRequest, CancelTaskResponse, CreateTaskRequest, CreateTaskResponse,
+    GetFunctionRequest, GetFunctionResponse, GetFunctionUsageStatsRequest,
+    GetFunctionUsageStatsResponse, GetTaskRequest, GetTaskResponse, InvokeTaskRequest,
+    InvokeTaskResponse, RegisterFunctionRequest, RegisterFunctionRequestBuilder,
     RegisterFunctionResponse, RegisterInputFileRequest, RegisterInputFileResponse,
     RegisterOutputFileRequest, RegisterOutputFileResponse,
 };
 pub use teaclave_types::{
-    EnclaveInfo, Executor, FileCrypto, FunctionInput, FunctionOutput, TaskResult,
+    EnclaveInfo, Executor, FileCrypto, FunctionArgument, FunctionInput, FunctionOutput,
+    FunctionUsage, TaskResult,
 };
 
 pub mod bindings;
@@ -56,6 +59,13 @@ impl AuthenticationClient {
         Self { api_client }
     }
 
+    pub fn set_credential(&mut self, id: &str, token: &str) {
+        let mut metadata = HashMap::new();
+        metadata.insert("id".to_string(), id.to_string());
+        metadata.insert("token".to_string(), token.to_string());
+        self.api_client.set_metadata(metadata);
+    }
+
     pub fn user_register_with_request(
         &mut self,
         request: UserRegisterRequest,
@@ -75,8 +85,14 @@ impl AuthenticationClient {
         Ok(serialized_response)
     }
 
-    pub fn user_register(&mut self, user_id: &str, user_password: &str) -> Result<()> {
-        let request = UserRegisterRequest::new(user_id, user_password);
+    pub fn user_register(
+        &mut self,
+        user_id: &str,
+        user_password: &str,
+        role: &str,
+        attribute: &str,
+    ) -> Result<()> {
+        let request = UserRegisterRequest::new(user_id, user_password, role, attribute);
         let _response = self.user_register_with_request(request)?;
 
         Ok(())
@@ -190,33 +206,42 @@ impl FrontendClient {
         Ok(response)
     }
 
+    #[allow(clippy::too_many_arguments)]
     pub fn register_function(
         &mut self,
         name: &str,
         description: &str,
         executor_type: &str,
         payload: Option<&[u8]>,
-        arguments: Option<&[&str]>,
+        arguments: Option<Vec<FunctionArgument>>,
         inputs: Option<Vec<FunctionInput>>,
         outputs: Option<Vec<FunctionOutput>>,
+        usage_quota: Option<i32>,
     ) -> Result<String> {
         let executor_type = executor_type.try_into()?;
-        let mut request = RegisterFunctionRequest::new()
+        let mut builder = RegisterFunctionRequestBuilder::new()
             .name(name)
             .description(description)
             .executor_type(executor_type);
+
         if let Some(payload) = payload {
-            request = request.payload(payload.into());
+            builder = builder.payload(payload.into());
         }
         if let Some(arguments) = arguments {
-            request = request.arguments(arguments);
+            builder = builder.arguments(arguments);
         }
         if let Some(inputs) = inputs {
-            request = request.inputs(inputs);
+            builder = builder.inputs(inputs);
         }
         if let Some(outputs) = outputs {
-            request = request.outputs(outputs);
+            builder = builder.outputs(outputs);
         }
+        if let Some(usage_quota) = usage_quota {
+            let usage_quota = (usage_quota >= 0).then_some(usage_quota);
+            builder = builder.usage_quota(usage_quota)
+        }
+
+        let request = builder.build();
         let response = self.register_function_with_request(request)?;
 
         Ok(response.function_id.to_string())
@@ -248,6 +273,37 @@ impl FrontendClient {
         Ok(response)
     }
 
+    pub fn get_function_usage_stats_serialized(
+        &mut self,
+        serialized_request: &str,
+    ) -> Result<String> {
+        let request: frontend_proto::GetFunctionUsageStatsRequest =
+            serde_json::from_str(serialized_request)?;
+        let response: frontend_proto::GetFunctionUsageStatsResponse = self
+            .get_function_usage_stats_with_request(request.try_into()?)?
+            .into();
+        let serialized_response = serde_json::to_string(&response)?;
+
+        Ok(serialized_response)
+    }
+
+    pub fn get_function_usage_stats(&mut self, function_id: &str) -> Result<i32> {
+        let function_id = function_id.try_into()?;
+        let request = GetFunctionUsageStatsRequest::new(function_id);
+        let response = self.get_function_usage_stats_with_request(request)?;
+
+        Ok(response.current_usage)
+    }
+
+    pub fn get_function_usage_stats_with_request(
+        &mut self,
+        request: GetFunctionUsageStatsRequest,
+    ) -> Result<GetFunctionUsageStatsResponse> {
+        let response = self.api_client.get_function_usage_stats(request)?;
+
+        Ok(response)
+    }
+
     pub fn register_input_file_with_request(
         &mut self,
         request: RegisterInputFileRequest,
@@ -477,17 +533,48 @@ impl FrontendClient {
         Ok(serialized_response)
     }
 
-    pub fn get_task_result(&mut self, task_id: &str) -> Result<Vec<u8>> {
+    pub fn get_task_result(&mut self, task_id: &str) -> Result<(Vec<u8>, Vec<String>)> {
         loop {
             let request = GetTaskRequest::new(task_id.try_into()?);
             let response = self.get_task_with_request(request)?;
-            if let TaskResult::Ok(task_outputs) = response.result {
-                return Ok(task_outputs.return_value);
+            match response.result {
+                TaskResult::NotReady => {
+                    std::thread::sleep(std::time::Duration::from_secs(1));
+                }
+                TaskResult::Ok(task_outputs) => {
+                    return Ok((task_outputs.return_value, task_outputs.log));
+                }
+                TaskResult::Err(task_error) => {
+                    return Err(anyhow::anyhow!(task_error.reason));
+                }
             }
-            let one_second = std::time::Duration::from_secs(1);
-            std::thread::sleep(one_second);
         }
     }
+
+    pub fn cancel_task_with_request(
+        &mut self,
+        request: CancelTaskRequest,
+    ) -> Result<CancelTaskResponse> {
+        let response = self.api_client.cancel_task(request)?;
+
+        Ok(response)
+    }
+
+    pub fn cancel_task_serialized(&mut self, serialized_request: &str) -> Result<String> {
+        let request: frontend_proto::CancelTaskRequest = serde_json::from_str(serialized_request)?;
+        let response: frontend_proto::CancelTaskResponse =
+            self.cancel_task_with_request(request.try_into()?)?.into();
+        let serialized_response = serde_json::to_string(&response)?;
+
+        Ok(serialized_response)
+    }
+
+    pub fn cancel_task(&mut self, task_id: &str) -> Result<()> {
+        let request = CancelTaskRequest::new(task_id.try_into()?);
+        let _ = self.cancel_task_with_request(request)?;
+
+        Ok(())
+    }
 }
 
 #[cfg(test)]
@@ -504,6 +591,8 @@ mod tests {
     const AS_ROOT_CA_CERT_PATH: &str = "../../keys/ias_root_ca_cert.pem";
     const USER_ID: &str = "rust_client_sdk_test_user";
     const USER_PASSWORD: &str = "test_password";
+    const ADMIN_ID: &str = "admin";
+    const ADMIN_PASSWORD: &str = "teaclave";
 
     #[test]
     fn test_authentication_service() {
@@ -513,7 +602,9 @@ mod tests {
         let mut client =
             AuthenticationService::connect("localhost:7776", &enclave_info, &as_root_ca_cert)
                 .unwrap();
-        let _ = client.user_register(USER_ID, USER_PASSWORD);
+        let token = client.user_login(ADMIN_ID, ADMIN_PASSWORD).unwrap();
+        client.set_credential(ADMIN_ID, &token);
+        let _ = client.user_register(USER_ID, USER_PASSWORD, "PlatformAdmin", "");
         client.user_login(USER_ID, USER_PASSWORD).unwrap();
     }
 
@@ -525,7 +616,9 @@ mod tests {
         let mut client =
             AuthenticationService::connect("localhost:7776", &enclave_info, &as_root_ca_cert)
                 .unwrap();
-        let _ = client.user_register(USER_ID, USER_PASSWORD);
+        let token = client.user_login(ADMIN_ID, ADMIN_PASSWORD).unwrap();
+        client.set_credential(ADMIN_ID, &token);
+        let _ = client.user_register(USER_ID, USER_PASSWORD, "PlatformAdmin", "");
         let token = client.user_login(USER_ID, USER_PASSWORD).unwrap();
 
         let mut client =
@@ -537,13 +630,47 @@ mod tests {
                 "An native echo function.",
                 "builtin",
                 None,
-                Some(&["message"]),
+                Some(vec![FunctionArgument::new("message", "", true)]),
                 None,
                 None,
+                Some(2),
             )
             .unwrap();
         let _ = client.get_function(&function_id).unwrap();
         let function_arguments = hashmap!("message" => "Hello, Teaclave!");
+        let task_id = client
+            .create_task(
+                &function_id,
+                Some(function_arguments.clone()),
+                "builtin",
+                None,
+                None,
+            )
+            .unwrap();
+
+        let _ = client.invoke_task(&task_id).unwrap();
+        let (result, log) = client.get_task_result(&task_id).unwrap();
+        let usage_number = client.get_function_usage_stats(&function_id).unwrap();
+        assert_eq!(result, b"Hello, Teaclave!");
+        assert!(log.is_empty());
+        assert_eq!(1, usage_number);
+
+        let task_id = client
+            .create_task(
+                &function_id,
+                Some(function_arguments.clone()),
+                "builtin",
+                None,
+                None,
+            )
+            .unwrap();
+        let _ = client.invoke_task(&task_id).unwrap();
+        let (result, log) = client.get_task_result(&task_id).unwrap();
+        let usage_number = client.get_function_usage_stats(&function_id).unwrap();
+        assert_eq!(result, b"Hello, Teaclave!");
+        assert!(log.is_empty());
+        assert_eq!(2, usage_number);
+
         let task_id = client
             .create_task(
                 &function_id,
@@ -553,10 +680,7 @@ mod tests {
                 None,
             )
             .unwrap();
-
-        let _ = client.invoke_task(&task_id).unwrap();
-        let result = client.get_task_result(&task_id).unwrap();
-        assert_eq!(result, b"Hello, Teaclave!")
+        assert!(client.invoke_task(&task_id).is_err());
     }
 
     #[test]
@@ -567,7 +691,9 @@ mod tests {
         let mut client =
             AuthenticationService::connect("localhost:7776", &enclave_info, &as_root_ca_cert)
                 .unwrap();
-        let _ = client.user_register(USER_ID, USER_PASSWORD);
+        let token = client.user_login(ADMIN_ID, ADMIN_PASSWORD).unwrap();
+        client.set_credential(ADMIN_ID, &token);
+        let _ = client.user_register(USER_ID, USER_PASSWORD, "PlatformAdmin", "");
         let token = client.user_login(USER_ID, USER_PASSWORD).unwrap();
 
         let mut client =
@@ -609,7 +735,9 @@ mod tests {
         let mut client =
             AuthenticationService::connect("localhost:7776", &enclave_info, &as_root_ca_cert)
                 .unwrap();
-        let _ = client.user_register(USER_ID, USER_PASSWORD);
+        let token = client.user_login(ADMIN_ID, ADMIN_PASSWORD).unwrap();
+        client.set_credential(ADMIN_ID, &token);
+        let _ = client.user_register(USER_ID, USER_PASSWORD, "PlatformAdmin", "");
         let token = client.user_login(USER_ID, USER_PASSWORD).unwrap();
 
         let mut client =
@@ -645,7 +773,9 @@ mod tests {
         let mut client =
             AuthenticationService::connect("localhost:7776", &enclave_info, &as_root_ca_cert)
                 .unwrap();
-        let _ = client.user_register(USER_ID, USER_PASSWORD);
+        let token = client.user_login(ADMIN_ID, ADMIN_PASSWORD).unwrap();
+        client.set_credential(ADMIN_ID, &token);
+        let _ = client.user_register(USER_ID, USER_PASSWORD, "PlatformAdmin", "");
         let token = client.user_login(USER_ID, USER_PASSWORD).unwrap();
 
         let mut client =
@@ -682,7 +812,9 @@ mod tests {
         let mut client =
             AuthenticationService::connect("localhost:7776", &enclave_info, &as_root_ca_cert)
                 .unwrap();
-        let _ = client.user_register(USER_ID, USER_PASSWORD);
+        let token = client.user_login(ADMIN_ID, ADMIN_PASSWORD).unwrap();
+        client.set_credential(ADMIN_ID, &token);
+        let _ = client.user_register(USER_ID, USER_PASSWORD, "PlatformAdmin", "");
         let token = client.user_login(USER_ID, USER_PASSWORD).unwrap();
 
         let mut client =
@@ -710,4 +842,42 @@ mod tests {
         client.assign_data(&task_id, None, Some(outputs)).unwrap();
         client.approve_task(&task_id).unwrap();
     }
+
+    #[test]
+    fn test_cancel_task() {
+        let enclave_info = EnclaveInfo::from_file(ENCLAVE_INFO_PATH).unwrap();
+        let bytes = fs::read(AS_ROOT_CA_CERT_PATH).unwrap();
+        let as_root_ca_cert = pem::parse(bytes).unwrap().contents;
+        let mut client =
+            AuthenticationService::connect("localhost:7776", &enclave_info, &as_root_ca_cert)
+                .unwrap();
+        let token = client.user_login(ADMIN_ID, ADMIN_PASSWORD).unwrap();
+        client.set_credential(ADMIN_ID, &token);
+        let _ = client.user_register(USER_ID, USER_PASSWORD, "PlatformAdmin", "");
+        let token = client.user_login(USER_ID, USER_PASSWORD).unwrap();
+
+        let mut client =
+            FrontendService::connect("localhost:7777", &enclave_info, &as_root_ca_cert).unwrap();
+        client.set_credential(USER_ID, &token);
+        let function_id = "function-00000000-0000-0000-0000-000000000002";
+        let function_arguments = hashmap!("arg1" => "arg1_value");
+        let outputs_ownership = hashmap!("output" => vec![USER_ID.to_string()]);
+        let task_id = client
+            .create_task(
+                &function_id,
+                Some(function_arguments),
+                "mesapy",
+                None,
+                Some(outputs_ownership),
+            )
+            .unwrap();
+
+        print!("SDK DEBUG: task created");
+
+        let result = client.cancel_task(&task_id);
+        print!("SDK DEBUG: canceled, {:?}", result);
+
+        let task = client.get_task_result(&task_id);
+        assert!(task.is_err());
+    }
 }
